from qulacs_core.observable import *
from scipy.sparse import csr_matrix, kron
import numpy as np

sigmaz = csr_matrix([[1, 0], [0, -1]])
sigmay = csr_matrix([[0, -1j], [1j, 0]])
sigmax = csr_matrix([[0, 1], [1, 0]])
sigmai = csr_matrix([[1, 0], [0, 1]])

sigma_list = [sigmai, sigmax, sigmay, sigmaz]


def _kron_n(*ops):
    """
    takes tensor product of given scipy matrix
    Args:
        ops (:class:`list`) 
    """
    if len(ops) == 2:
        return kron(ops[0], ops[1])
    else:
        return kron(_kron_n(*ops[:-1]), ops[-1])


def _get_matrix(obs):
    """
    returns matrix of an observable
    Args:
        obs (qulacs_core.Observable)
    Return:
        scipy.sparse.csr_matrix
    """
    n_terms = obs.get_term_count()
    n_qubits = obs.get_qubit_count()
    result = csr_matrix((2**n_qubits, 2**n_qubits), dtype=np.complex128)
    for i in range(n_terms):
        pauli = obs.get_term(i)
        pauli_id_list = pauli.get_pauli_id_list()
        pauli_target_list = pauli.get_index_list()
        pauli_string = [sigmai for q in range(n_qubits)]
        for j, target in enumerate(pauli_target_list):
            pauli_string[target] = sigma_list[pauli_id_list[j]]
        result += pauli.get_coef()*_kron_n(*(pauli_string[::-1]))
    return result



import qulacs_core

class QuantumCircuitOptimizer:
    def __init__(self) -> None: ...
    def merge_all(self, circuit: qulacs_core.QuantumCircuit) -> qulacs_core.QuantumGateMatrix: ...
    def optimize(self, circuit: qulacs_core.QuantumCircuit, block_size: int) -> None: ...
    def optimize_light(self, circuit: qulacs_core.QuantumCircuit) -> None: ...



import numpy as np
from scipy.sparse import csr_matrix
from qulacs import QuantumCircuit, QuantumState
from qulacs.gate import X, T, H, CNOT, TOFFOLI, DenseMatrix, PauliRotation, SparseMatrix, DiagonalMatrix, ReversibleBoolean

import pytest

nqubits_list = range(4,26)

def bench_gate(benchmark, nqubits, g):
    st = QuantumState(nqubits)
    benchmark(g.update_quantum_state, st)

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_X(benchmark, nqubits):
    benchmark.group = "X"
    bench_gate(benchmark, nqubits, X(3))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_CNOT(benchmark, nqubits):
    benchmark.group = "CNOT"
    bench_gate(benchmark, nqubits, CNOT(2, 3))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_TOFFOLI(benchmark, nqubits):
    benchmark.group = "TOFFOLI"
    bench_gate(benchmark, nqubits, TOFFOLI(1, 2, 3))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_CCCNOT(benchmark, nqubits):
    benchmark.group = "CCCNOT"
    g = DenseMatrix([3], [[0,1],[1,0]])
    g.add_control_qubit(2,1)
    g.add_control_qubit(1,1)
    g.add_control_qubit(0,1)
    bench_gate(benchmark, nqubits, g)

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Dense1Q(benchmark, nqubits):
    benchmark.group = "DenseMatrix1Q"
    bench_gate(benchmark, nqubits, DenseMatrix([3], np.eye(2)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Dense2Q(benchmark, nqubits):
    benchmark.group = "DenseMatrix2Q"
    bench_gate(benchmark, nqubits, DenseMatrix([2,3], np.eye(4)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Dense3Q(benchmark, nqubits):
    benchmark.group = "DenseMatrix3Q"
    bench_gate(benchmark, nqubits, DenseMatrix([1,2,3], np.eye(8)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Dense4Q(benchmark, nqubits):
    benchmark.group = "DenseMatrix4Q"
    bench_gate(benchmark, nqubits, DenseMatrix([0,1,2,3], np.eye(16)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Diagonal1QSp(benchmark, nqubits):
    benchmark.group = "DiagonalMatrix1QSp"
    bench_gate(benchmark, nqubits, DiagonalMatrix([0], np.ones(2)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Diagonal1Q(benchmark, nqubits):
    benchmark.group = "DiagonalMatrix1Q"
    bench_gate(benchmark, nqubits, DiagonalMatrix([3], np.ones(2)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Diagonal2Q(benchmark, nqubits):
    benchmark.group = "DiagonalMatrix2Q"
    bench_gate(benchmark, nqubits, DiagonalMatrix([2,3], np.ones(4)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Diagonal3Q(benchmark, nqubits):
    benchmark.group = "DiagonalMatrix3Q"
    bench_gate(benchmark, nqubits, DiagonalMatrix([1,2,3], np.ones(8)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Diagonal4Q(benchmark, nqubits):
    benchmark.group = "DiagonalMatrix4Q"
    bench_gate(benchmark, nqubits, DiagonalMatrix([0,1,2,3], np.ones(16)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Sparse1Q(benchmark, nqubits):
    benchmark.group = "SparseMatrix1Q"
    sparse_matrix = csr_matrix(([1], ([0],[0])), shape=(2,2), dtype=complex)
    bench_gate(benchmark, nqubits, SparseMatrix([3], sparse_matrix))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Sparse2Q(benchmark, nqubits):
    benchmark.group = "SparseMatrix2Q"
    sparse_matrix = csr_matrix(([1], ([0],[0])), shape=(4,4), dtype=complex)
    bench_gate(benchmark, nqubits, SparseMatrix([2,3], sparse_matrix))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Sparse3Q(benchmark, nqubits):
    benchmark.group = "SparseMatrix3Q"
    sparse_matrix = csr_matrix(([1], ([0],[0])), shape=(8,8), dtype=complex)
    bench_gate(benchmark, nqubits, SparseMatrix([1,2,3], sparse_matrix))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Sparse4Q(benchmark, nqubits):
    benchmark.group = "SparseMatrix4Q"
    sparse_matrix = csr_matrix(([1], ([0],[0])), shape=(16,16), dtype=complex)
    bench_gate(benchmark, nqubits, SparseMatrix([0,1,2,3], sparse_matrix))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Permutation1Q(benchmark, nqubits):
    benchmark.group = "Permutation1Q"
    def rev(index, dim):
        return (index+1)%dim
    bench_gate(benchmark, nqubits, ReversibleBoolean([3], rev))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Permutation2Q(benchmark, nqubits):
    benchmark.group = "Permutation2Q"
    def rev(index, dim):
        return (index+1)%dim
    bench_gate(benchmark, nqubits, ReversibleBoolean([2,3], rev))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Permutation3Q(benchmark, nqubits):
    benchmark.group = "Permutation3Q"
    def rev(index, dim):
        return (index+1)%dim
    bench_gate(benchmark, nqubits, ReversibleBoolean([1,2,3], rev))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Permutation4Q(benchmark, nqubits):
    benchmark.group = "Permutation4Q"
    def rev(index, dim):
        return (index+1)%dim
    bench_gate(benchmark, nqubits, ReversibleBoolean([0,1,2,3], rev))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_PauliRotation1Q(benchmark, nqubits):
    benchmark.group = "PauliRotation1Q"
    bench_gate(benchmark, nqubits, PauliRotation([3], [1], 0.1))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_PauliRotation2Q(benchmark, nqubits):
    benchmark.group = "PauliRotation2Q"
    bench_gate(benchmark, nqubits, PauliRotation([2,3], [1,1], 0.1))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_PauliRotation3Q(benchmark, nqubits):
    benchmark.group = "PauliRotation3Q"
    bench_gate(benchmark, nqubits, PauliRotation([1,2,3], [1,1,1], 0.1))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_PauliRotation4Q(benchmark, nqubits):
    benchmark.group = "PauliRotation4Q"
    bench_gate(benchmark, nqubits, PauliRotation([0,1,2,3], [1,1,1,1], 0.1))



from qulacs_core import *

import qulacs.observable._get_matrix
from qulacs._version import __version__, __version_tuple__

Observable.get_matrix = \
    lambda obs: qulacs.observable._get_matrix._get_matrix(obs)

GeneralQuantumOperator.get_matrix = \
    lambda obs: qulacs.observable._get_matrix._get_matrix(obs)



import textwrap

project = "Qulacs"

language = 'ja'
locale_dirs = ['locale/']

# The `extensions` list should already be in here from `sphinx-quickstart`
extensions = [
    # there may be others here already, e.g. 'sphinx.ext.mathjax'
    # there may be others here already, e.g. 'sphinx.ext.mathjax'
    'breathe',
    'exhale',
    'myst_parser',
    'nbsphinx',
    'sphinx.ext.mathjax',
    'sphinx_copybutton',
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'sphinx.ext.graphviz',
    'sphinx.ext.inheritance_diagram',
]

exclude_patterns = ['_build', '**.ipynb_checkpoints']
nbsphinx_allow_errors = True
myst_enable_extensions = [
    "dollarmath",
]


# source files for shpinx
source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}

# Setup the breathe extension
breathe_projects = {
    "Docs": "./xml"
}

breathe_default_project = "Docs"

# Setup the exhale extension
exhale_args = {
    # These arguments are required
    "containmentFolder":     "./api",
    "rootFileName":          "cpp_library_root.rst",
    "rootFileTitle":         "C++ APIリファレンス",
    "doxygenStripFromPath":  "..",
    # Suggested optional arguments
    "createTreeView":        True,
    # TIP: if using the sphinx-bootstrap-theme, you need
    # "treeViewIsBootstrap": True,
    "exhaleExecutesDoxygen": True,
    "exhaleDoxygenStdin": \
        "INPUT = ../../../src/cppsim ../../../src/vqcsim \n \
        OUTPUT_LANGUAGE = Japanese-en \n \
        FILE_PATTERNS          = *.hpp \n \
        WARN_IF_UNDOCUMENTED   = NO \n \
        ENABLE_PREPROCESSING   = YES \n \
        MACRO_EXPANSION        = YES \n \
        PREDEFINED             += __attribute__(x)= \n \
        PREDEFINED             += DllExport= \n \
        GENERATE_LEGEND        = YES \n \
        GRAPHICAL_HIERARCHY    = YES \n \
        CLASS_GRAPH            = YES \n \
        HIDE_UNDOC_RELATIONS   = YES\n \
        CLASS_DIAGRAMS         = YES\n"
}

# Tell sphinx what the primary language being documented is.
#primary_domain = 'cpp'

# Tell sphinx what the pygments highlight language should be.
#highlight_language = 'cpp'

#

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
import os

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


html_theme_options = {
    # 'canonical_url': '',
    # 'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    'logo_only': False,
    # 'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    # 'vcs_pageview_mode': '',
    'style_nav_header_background': '#004659',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 2,
    'includehidden': True,
    'titles_only': False
}

templates_path = ["_templates"]
html_static_path = ['_static']

copyright = '2018 Qulacs Authors'
version = '0.5.4'



from typing import Callable, List

from typing import overload
import numpy
import qulacs_core
import scipy.sparse

@overload
def Adaptive(gate: qulacs_core.QuantumGateBase, condition: Callable[[List[int]],bool]) -> qulacs_core.QuantumGateBase: ...
@overload
def Adaptive(gate: qulacs_core.QuantumGateBase, condition: Callable[[List[int],int],bool], id: int) -> qulacs_core.QuantumGateBase: ...
def AmplitudeDampingNoise(index: int, prob: float) -> qulacs_core.QuantumGate_CPTP: ...
def BitFlipNoise(index: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic: ...
def CNOT(control: int, target: int) -> qulacs_core.ClsOneControlOneTargetGate: ...
def CP(kraus_list: List[qulacs_core.QuantumGateBase], state_normalize: bool, probability_normalize: bool, assign_zero_if_not_matched: bool) -> qulacs_core.QuantumGateBase: ...
def CPTP(kraus_list: List[qulacs_core.QuantumGateBase]) -> qulacs_core.QuantumGateBase: ...
def CZ(control: int, target: int) -> qulacs_core.ClsOneControlOneTargetGate: ...
@overload
def DenseMatrix(index: int, matrix: numpy.ndarray[numpy.complex128[m,n]]) -> qulacs_core.QuantumGateMatrix: ...
@overload
def DenseMatrix(index_list: List[int], matrix: numpy.ndarray[numpy.complex128[m,n]]) -> qulacs_core.QuantumGateMatrix: ...
def DephasingNoise(index: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic: ...
def DepolarizingNoise(index: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic: ...
def DiagonalMatrix(index_list: List[int], diagonal_element: numpy.ndarray[numpy.complex128[m,1]]) -> qulacs_core.QuantumGateDiagonalMatrix: ...
def FREDKIN(control: int, target1: int, target2: int) -> qulacs_core.QuantumGateMatrix: ...
def H(index: int) -> qulacs_core.ClsOneQubitGate: ...
def Identity(index: int) -> qulacs_core.ClsOneQubitGate: ...
def IndependentXZNoise(index: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic: ...
def Instrument(kraus_list: List[qulacs_core.QuantumGateBase], register: int) -> qulacs_core.QuantumGateBase: ...
def Measurement(index: int, register: int) -> qulacs_core.QuantumGate_CPTP: ...
def NoisyEvolution(hamiltonian: qulacs_core.Observable, c_ops: List[qulacs_core.GeneralQuantumOperator], time: float, dt: float) -> qulacs_core.ClsNoisyEvolution: ...
def NoisyEvolution_fast(hamiltonian: qulacs_core.Observable, c_ops: List[qulacs_core.GeneralQuantumOperator], time: float) -> qulacs_core.ClsNoisyEvolution_fast: ...
def P0(index: int) -> qulacs_core.ClsOneQubitGate: ...
def P1(index: int) -> qulacs_core.ClsOneQubitGate: ...
def ParametricPauliRotation(index_list: List[int], pauli_ids: List[int], angle: float) -> qulacs_core.QuantumGate_SingleParameter: ...
def ParametricRX(index: int, angle: float) -> qulacs_core.QuantumGate_SingleParameter: ...
def ParametricRY(index: int, angle: float) -> qulacs_core.QuantumGate_SingleParameter: ...
def ParametricRZ(index: int, angle: float) -> qulacs_core.QuantumGate_SingleParameter: ...
def Pauli(index_list: List[int], pauli_ids: List[int]) -> qulacs_core.ClsPauliGate: ...
def PauliRotation(index_list: List[int], pauli_ids: List[int], angle: float) -> qulacs_core.ClsPauliRotationGate: ...
def Probabilistic(prob_list: List[float], gate_list: List[qulacs_core.QuantumGateBase]) -> qulacs_core.QuantumGateBase: ...
def ProbabilisticInstrument(prob_list: List[float], gate_list: List[qulacs_core.QuantumGateBase], register: int) -> qulacs_core.QuantumGateBase: ...
def RX(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
def RY(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
def RZ(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
@overload
def RandomUnitary(index_list: List[int]) -> qulacs_core.QuantumGateMatrix: ...
@overload
def RandomUnitary(index_list: List[int], seed: int) -> qulacs_core.QuantumGateMatrix: ...
def ReversibleBoolean(index_list: List[int], func: Callable[[int,int],int]) -> qulacs_core.ClsReversibleBooleanGate: ...
def RotInvX(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
def RotInvY(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
def RotInvZ(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
def RotX(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
def RotY(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
def RotZ(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate: ...
def S(index: int) -> qulacs_core.ClsOneQubitGate: ...
def SWAP(target1: int, target2: int) -> qulacs_core.ClsTwoQubitGate: ...
def Sdag(index: int) -> qulacs_core.ClsOneQubitGate: ...
def SparseMatrix(index_list: List[int], matrix: scipy.sparse.csc_matrix[numpy.complex128]) -> qulacs_core.QuantumGateSparseMatrix: ...
def StateReflection(state: qulacs_core.QuantumState) -> qulacs_core.ClsStateReflectionGate: ...
def T(index: int) -> qulacs_core.ClsOneQubitGate: ...
def TOFFOLI(control1: int, control2: int, target: int) -> qulacs_core.QuantumGateMatrix: ...
def Tdag(index: int) -> qulacs_core.ClsOneQubitGate: ...
def TwoQubitDepolarizingNoise(index1: int, index2: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic: ...
def U1(index: int, lambda: float) -> qulacs_core.QuantumGateMatrix: ...
def U2(index: int, phi: float, lambda: float) -> qulacs_core.QuantumGateMatrix: ...
def U3(index: int, theta: float, phi: float, lambda: float) -> qulacs_core.QuantumGateMatrix: ...
def X(index: int) -> qulacs_core.ClsOneQubitGate: ...
def Y(index: int) -> qulacs_core.ClsOneQubitGate: ...
def Z(index: int) -> qulacs_core.ClsOneQubitGate: ...
@overload
def add(gate1: qulacs_core.QuantumGateBase, gate2: qulacs_core.QuantumGateBase) -> qulacs_core.QuantumGateMatrix: ...
@overload
def add(gate_list: List[qulacs_core.QuantumGateBase]) -> qulacs_core.QuantumGateMatrix: ...
@overload
def merge(gate1: qulacs_core.QuantumGateBase, gate2: qulacs_core.QuantumGateBase) -> qulacs_core.QuantumGateMatrix: ...
@overload
def merge(gate_list: List[qulacs_core.QuantumGateBase]) -> qulacs_core.QuantumGateMatrix: ...
def sqrtX(index: int) -> qulacs_core.ClsOneQubitGate: ...
def sqrtXdag(index: int) -> qulacs_core.ClsOneQubitGate: ...
def sqrtY(index: int) -> qulacs_core.ClsOneQubitGate: ...
def sqrtYdag(index: int) -> qulacs_core.ClsOneQubitGate: ...
def to_matrix_gate(gate: qulacs_core.QuantumGateBase) -> qulacs_core.QuantumGateMatrix: ...



from __future__ import annotations
import qulacs_core.observable
import typing
import qulacs_core

__all__ = [
    "create_observable_from_openfermion_file",
    "create_observable_from_openfermion_text",
    "create_split_observable"
]


def create_observable_from_openfermion_file(file_path: str) -> qulacs_core.Observable:
    """
    Create GeneralQuantumOperator from openfermion file
    """
def create_observable_from_openfermion_text(text: str) -> qulacs_core.Observable:
    """
    Create GeneralQuantumOperator from openfermion text
    """
def create_split_observable(arg0: str) -> typing.Tuple[qulacs_core.Observable, qulacs_core.Observable]:
    pass



from __future__ import annotations
import qulacs_core.quantum_operator
import typing
import qulacs_core

__all__ = [
    "create_quantum_operator_from_openfermion_file",
    "create_quantum_operator_from_openfermion_text",
    "create_split_quantum_operator"
]


def create_quantum_operator_from_openfermion_file(arg0: str) -> qulacs_core.GeneralQuantumOperator:
    pass
def create_quantum_operator_from_openfermion_text(arg0: str) -> qulacs_core.GeneralQuantumOperator:
    pass
def create_split_quantum_operator(arg0: str) -> typing.Tuple[qulacs_core.GeneralQuantumOperator, qulacs_core.GeneralQuantumOperator]:
    pass



from typing import List

from typing import overload
import qulacs_core

def drop_qubit(state: qulacs_core.QuantumState, target: List[int], projection: List[int]) -> qulacs_core.QuantumState: ...
def inner_product(state_bra: qulacs_core.QuantumState, state_ket: qulacs_core.QuantumState) -> complex: ...
def make_mixture(prob1: complex, state1: qulacs_core.QuantumStateBase, prob2: complex, state2: qulacs_core.QuantumStateBase) -> qulacs_core.DensityMatrix: ...
def make_superposition(coef1: complex, state1: qulacs_core.QuantumState, coef2: complex, state2: qulacs_core.QuantumState) -> qulacs_core.QuantumState: ...
@overload
def partial_trace(state: qulacs_core.QuantumState, target_traceout: List[int]) -> qulacs_core.DensityMatrix: ...
@overload
def partial_trace(state: qulacs_core.DensityMatrix, target_traceout: List[int]) -> qulacs_core.DensityMatrix: ...
@overload
def permutate_qubit(state: qulacs_core.QuantumState, qubit_order: List[int]) -> qulacs_core.QuantumState: ...
@overload
def permutate_qubit(state: qulacs_core.DensityMatrix, qubit_order: List[int]) -> qulacs_core.DensityMatrix: ...
@overload
def tensor_product(state_left: qulacs_core.QuantumState, state_right: qulacs_core.QuantumState) -> qulacs_core.QuantumState: ...
@overload
def tensor_product(state_left: qulacs_core.DensityMatrix, state_right: qulacs_core.DensityMatrix) -> qulacs_core.DensityMatrix: ...



from __future__ import annotations
import qulacs_core.gate
import typing
import numpy
import qulacs_core
import scipy.sparse
_Shape = typing.Tuple[int, ...]

__all__ = [
    "Adaptive",
    "AmplitudeDampingNoise",
    "BitFlipNoise",
    "CNOT",
    "CP",
    "CPTP",
    "CZ",
    "DenseMatrix",
    "DephasingNoise",
    "DepolarizingNoise",
    "DiagonalMatrix",
    "FREDKIN",
    "H",
    "Identity",
    "IndependentXZNoise",
    "Instrument",
    "Measurement",
    "NoisyEvolution",
    "NoisyEvolution_fast",
    "P0",
    "P1",
    "ParametricPauliRotation",
    "ParametricRX",
    "ParametricRY",
    "ParametricRZ",
    "Pauli",
    "PauliRotation",
    "Probabilistic",
    "ProbabilisticInstrument",
    "RX",
    "RY",
    "RZ",
    "RandomUnitary",
    "ReversibleBoolean",
    "RotInvX",
    "RotInvY",
    "RotInvZ",
    "RotX",
    "RotY",
    "RotZ",
    "S",
    "SWAP",
    "Sdag",
    "SparseMatrix",
    "StateReflection",
    "T",
    "TOFFOLI",
    "Tdag",
    "TwoQubitDepolarizingNoise",
    "U1",
    "U2",
    "U3",
    "X",
    "Y",
    "Z",
    "add",
    "merge",
    "sqrtX",
    "sqrtXdag",
    "sqrtY",
    "sqrtYdag",
    "to_matrix_gate"
]


@typing.overload
def Adaptive(gate: qulacs_core.QuantumGateBase, condition: typing.Callable[[typing.List[int], int], bool], id: int) -> qulacs_core.QuantumGateBase:
    """
    Create adaptive gate
    """
@typing.overload
def Adaptive(gate: qulacs_core.QuantumGateBase, condition: typing.Callable[[typing.List[int]], bool]) -> qulacs_core.QuantumGateBase:
    pass
def AmplitudeDampingNoise(index: int, prob: float) -> qulacs_core.QuantumGate_CPTP:
    """
    Create amplitude damping noise
    """
def BitFlipNoise(index: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic:
    """
    Create bit-flip noise
    """
def CNOT(control: int, target: int) -> qulacs_core.ClsOneControlOneTargetGate:
    """
    Create CNOT gate
    """
def CP(kraus_list: typing.List[qulacs_core.QuantumGateBase], state_normalize: bool, probability_normalize: bool, assign_zero_if_not_matched: bool) -> qulacs_core.QuantumGateBase:
    """
    Create completely-positive map
    """
def CPTP(kraus_list: typing.List[qulacs_core.QuantumGateBase]) -> qulacs_core.QuantumGateBase:
    """
    Create completely-positive trace preserving map
    """
def CZ(control: int, target: int) -> qulacs_core.ClsOneControlOneTargetGate:
    """
    Create CZ gate
    """
@typing.overload
def DenseMatrix(index: int, matrix: numpy.ndarray[numpy.complex128, _Shape[m, n]]) -> qulacs_core.QuantumGateMatrix:
    """
    Create dense matrix gate
    """
@typing.overload
def DenseMatrix(index_list: typing.List[int], matrix: numpy.ndarray[numpy.complex128, _Shape[m, n]]) -> qulacs_core.QuantumGateMatrix:
    pass
def DephasingNoise(index: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic:
    """
    Create dephasing noise
    """
def DepolarizingNoise(index: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic:
    """
    Create depolarizing noise
    """
def DiagonalMatrix(index_list: typing.List[int], diagonal_element: numpy.ndarray[numpy.complex128, _Shape[m, 1]]) -> qulacs_core.QuantumGateDiagonalMatrix:
    """
    Create diagonal matrix gate
    """
def FREDKIN(control: int, target1: int, target2: int) -> qulacs_core.QuantumGateMatrix:
    """
    Create FREDKIN gate
    """
def H(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create Hadamard gate
    """
def Identity(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create identity gate
    """
def IndependentXZNoise(index: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic:
    """
    Create independent XZ noise
    """
def Instrument(kraus_list: typing.List[qulacs_core.QuantumGateBase], register: int) -> qulacs_core.QuantumGateBase:
    """
    Create instruments
    """
def Measurement(index: int, register: int) -> qulacs_core.QuantumGate_CPTP:
    """
    Create measurement gate
    """
def NoisyEvolution(hamiltonian: qulacs_core.Observable, c_ops: typing.List[qulacs_core.GeneralQuantumOperator], time: float, dt: float) -> qulacs_core.ClsNoisyEvolution:
    """
    Create noisy evolution
    """
def NoisyEvolution_fast(hamiltonian: qulacs_core.Observable, c_ops: typing.List[qulacs_core.GeneralQuantumOperator], time: float) -> qulacs_core.ClsNoisyEvolution_fast:
    """
    Create noisy evolution fast version
    """
def P0(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create projection gate to |0> subspace
    """
def P1(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create projection gate to |1> subspace
    """
def ParametricPauliRotation(index_list: typing.List[int], pauli_ids: typing.List[int], angle: float) -> qulacs_core.QuantumGate_SingleParameter:
    """
    Create parametric multi-qubit Pauli rotation gate
    """
def ParametricRX(index: int, angle: float) -> qulacs_core.QuantumGate_SingleParameter:
    """
    Create parametric Pauli-X rotation gate
    """
def ParametricRY(index: int, angle: float) -> qulacs_core.QuantumGate_SingleParameter:
    """
    Create parametric Pauli-Y rotation gate
    """
def ParametricRZ(index: int, angle: float) -> qulacs_core.QuantumGate_SingleParameter:
    """
    Create parametric Pauli-Z rotation gate
    """
def Pauli(index_list: typing.List[int], pauli_ids: typing.List[int]) -> qulacs_core.ClsPauliGate:
    """
    Create multi-qubit Pauli gate
    """
def PauliRotation(index_list: typing.List[int], pauli_ids: typing.List[int], angle: float) -> qulacs_core.ClsPauliRotationGate:
    """
    Create multi-qubit Pauli rotation
    """
def Probabilistic(prob_list: typing.List[float], gate_list: typing.List[qulacs_core.QuantumGateBase]) -> qulacs_core.QuantumGateBase:
    """
    Create probabilistic gate
    """
def ProbabilisticInstrument(prob_list: typing.List[float], gate_list: typing.List[qulacs_core.QuantumGateBase], register: int) -> qulacs_core.QuantumGateBase:
    """
    Create probabilistic instrument gate
    """
def RX(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-X rotation gate
    """
def RY(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-Y rotation gate
    """
def RZ(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-Z rotation gate
    """
@typing.overload
def RandomUnitary(index_list: typing.List[int]) -> qulacs_core.QuantumGateMatrix:
    """
    Create random unitary gate
    """
@typing.overload
def RandomUnitary(index_list: typing.List[int], seed: int) -> qulacs_core.QuantumGateMatrix:
    pass
def ReversibleBoolean(index_list: typing.List[int], func: typing.Callable[[int, int], int]) -> qulacs_core.ClsReversibleBooleanGate:
    """
    Create reversible boolean gate
    """
def RotInvX(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-X rotation gate
    """
def RotInvY(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-Y rotation gate
    """
def RotInvZ(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-Z rotation gate
    """
def RotX(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-X rotation gate
    """
def RotY(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-Y rotation gate
    """
def RotZ(index: int, angle: float) -> qulacs_core.ClsOneQubitRotationGate:
    """
    Create Pauli-Z rotation gate
    """
def S(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create pi/4-phase gate
    """
def SWAP(target1: int, target2: int) -> qulacs_core.ClsTwoQubitGate:
    """
    Create SWAP gate
    """
def Sdag(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create adjoint of pi/4-phase gate
    """
def SparseMatrix(index_list: typing.List[int], matrix: scipy.sparse.csc_matrix[numpy.complex128]) -> qulacs_core.QuantumGateSparseMatrix:
    """
    Create sparse matrix gate
    """
def StateReflection(state: qulacs_core.QuantumState) -> qulacs_core.ClsStateReflectionGate:
    """
    Create state reflection gate
    """
def T(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create pi/8-phase gate
    """
def TOFFOLI(control1: int, control2: int, target: int) -> qulacs_core.QuantumGateMatrix:
    """
    Create TOFFOLI gate
    """
def Tdag(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create adjoint of pi/8-phase gate
    """
def TwoQubitDepolarizingNoise(index1: int, index2: int, prob: float) -> qulacs_core.QuantumGate_Probabilistic:
    """
    Create two-qubit depolarizing noise
    """
def U1(*args, **kwargs) -> typing.Any:
    """
    Create QASM U1 gate
    """
def U2(*args, **kwargs) -> typing.Any:
    """
    Create QASM U2 gate
    """
def U3(*args, **kwargs) -> typing.Any:
    """
    Create QASM U3 gate
    """
def X(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create Pauli-X gate
    """
def Y(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create Pauli-Y gate
    """
def Z(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create Pauli-Z gate
    """
@typing.overload
def add(gate1: qulacs_core.QuantumGateBase, gate2: qulacs_core.QuantumGateBase) -> qulacs_core.QuantumGateMatrix:
    """
    Add quantum gate matrices

    Add quantum gate matrices
    """
@typing.overload
def add(gate_list: typing.List[qulacs_core.QuantumGateBase]) -> qulacs_core.QuantumGateMatrix:
    pass
@typing.overload
def merge(gate1: qulacs_core.QuantumGateBase, gate2: qulacs_core.QuantumGateBase) -> qulacs_core.QuantumGateMatrix:
    """
    Merge two quantum gate or gate list
    """
@typing.overload
def merge(gate_list: typing.List[qulacs_core.QuantumGateBase]) -> qulacs_core.QuantumGateMatrix:
    pass
def sqrtX(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create pi/4 Pauli-X rotation gate
    """
def sqrtXdag(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create adjoint of pi/4 Pauli-X rotation gate
    """
def sqrtY(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create pi/4 Pauli-Y rotation gate
    """
def sqrtYdag(index: int) -> qulacs_core.ClsOneQubitGate:
    """
    Create adjoint of pi/4 Pauli-Y rotation gate
    """
def to_matrix_gate(gate: qulacs_core.QuantumGateBase) -> qulacs_core.QuantumGateMatrix:
    """
    Convert named gate to matrix gate
    """



from __future__ import annotations
import qulacs_core.state
import typing
import qulacs_core

__all__ = [
    "drop_qubit",
    "inner_product",
    "make_mixture",
    "make_superposition",
    "partial_trace",
    "permutate_qubit",
    "tensor_product"
]


def drop_qubit(state: qulacs_core.QuantumState, target: typing.List[int], projection: typing.List[int]) -> qulacs_core.QuantumState:
    """
    Drop qubits from state
    """
def inner_product(state_bra: qulacs_core.QuantumState, state_ket: qulacs_core.QuantumState) -> complex:
    """
    Get inner product
    """
def make_mixture(prob1: complex, state1: qulacs_core.QuantumStateBase, prob2: complex, state2: qulacs_core.QuantumStateBase) -> qulacs_core.DensityMatrix:
    """
    Create a mixed state
    """
def make_superposition(coef1: complex, state1: qulacs_core.QuantumState, coef2: complex, state2: qulacs_core.QuantumState) -> qulacs_core.QuantumState:
    """
    Create superposition of states
    """
@typing.overload
def partial_trace(state: qulacs_core.DensityMatrix, target_traceout: typing.List[int]) -> qulacs_core.DensityMatrix:
    """
    Take partial trace
    """
@typing.overload
def partial_trace(state: qulacs_core.QuantumState, target_traceout: typing.List[int]) -> qulacs_core.DensityMatrix:
    pass
@typing.overload
def permutate_qubit(state: qulacs_core.DensityMatrix, qubit_order: typing.List[int]) -> qulacs_core.DensityMatrix:
    """
    Permutate qubits from state
    """
@typing.overload
def permutate_qubit(state: qulacs_core.QuantumState, qubit_order: typing.List[int]) -> qulacs_core.QuantumState:
    pass
@typing.overload
def tensor_product(state_left: qulacs_core.DensityMatrix, state_right: qulacs_core.DensityMatrix) -> qulacs_core.DensityMatrix:
    """
    Get tensor product of states
    """
@typing.overload
def tensor_product(state_left: qulacs_core.QuantumState, state_right: qulacs_core.QuantumState) -> qulacs_core.QuantumState:
    pass



from qulacs.converter.qasm_converter import (convert_QASM_to_qulacs_circuit,
                                             convert_qulacs_circuit_to_QASM)

__all__ = ['convert_QASM_to_qulacs_circuit', 'convert_qulacs_circuit_to_QASM']



import numpy as np
from scipy.sparse import csr_matrix
from qulacs import QuantumCircuit, QuantumState
from qulacs.gate import DenseMatrix, CPTP, Probabilistic, Instrument, Measurement

import pytest

nqubits_list = range(4,26)

def bench_gate(benchmark, nqubits, g):
    st = QuantumState(nqubits)
    benchmark(g.update_quantum_state, st)

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Dense1Q(benchmark, nqubits):
    benchmark.group = "DenseMatrix1Q"
    bench_gate(benchmark, nqubits, DenseMatrix([3], np.eye(2)))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_CPTP1Q(benchmark, nqubits):
    benchmark.group = "CPTP1Q"
    g = DenseMatrix([3,], np.eye(2)/np.sqrt(2))
    cptp = CPTP([g,g])
    bench_gate(benchmark, nqubits, cptp)

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Instrument1Q(benchmark, nqubits):
    benchmark.group = "Instrument1Q"
    bench_gate(benchmark, nqubits, Measurement(3, 0))

@pytest.mark.parametrize('nqubits', nqubits_list)
def test_Probabilistic1Q(benchmark, nqubits):
    benchmark.group = "Probabilistic1Q"
    g = DenseMatrix([3,], np.eye(2)/np.sqrt(2))
    gate = Probabilistic([0.5, 0.5], [g,g])
    bench_gate(benchmark, nqubits, gate)




"""cppsim python interface"""
from __future__ import annotations
import qulacs_core
import typing
import numpy
_Shape = typing.Tuple[int, ...]

__all__ = [
    "CausalConeSimulator",
    "ClsNoisyEvolution",
    "ClsNoisyEvolution_fast",
    "ClsOneControlOneTargetGate",
    "ClsOneQubitGate",
    "ClsOneQubitRotationGate",
    "ClsPauliGate",
    "ClsPauliRotationGate",
    "ClsReversibleBooleanGate",
    "ClsStateReflectionGate",
    "ClsTwoQubitGate",
    "DensityMatrix",
    "GeneralQuantumOperator",
    "GradCalculator",
    "NoiseSimulator",
    "Observable",
    "ParametricQuantumCircuit",
    "PauliOperator",
    "QuantumCircuit",
    "QuantumCircuitSimulator",
    "QuantumGateBase",
    "QuantumGateDiagonalMatrix",
    "QuantumGateMatrix",
    "QuantumGateSparseMatrix",
    "QuantumGate_Adaptive",
    "QuantumGate_CP",
    "QuantumGate_CPTP",
    "QuantumGate_Probabilistic",
    "QuantumGate_SingleParameter",
    "QuantumState",
    "QuantumStateBase",
    "SimulationResult",
    "StateVector",
    "circuit",
    "gate",
    "observable",
    "quantum_operator",
    "state",
    "to_general_quantum_operator"
]


class CausalConeSimulator():
    def __init__(self, arg0: ParametricQuantumCircuit, arg1: Observable) -> None: 
        """
        Constructor
        """
    def build(self) -> None: 
        """
        Build
        """
    def get_circuit_list(self) -> typing.List[typing.List[ParametricQuantumCircuit]]: 
        """
        Return circuit_list
        """
    def get_coef_list(self) -> typing.List[complex]: 
        """
        Return coef_list
        """
    def get_expectation_value(self) -> complex: 
        """
        Return expectation_value
        """
    def get_pauli_operator_list(self) -> typing.List[typing.List[PauliOperator]]: 
        """
        Return pauli_operator_list
        """
    pass
class QuantumGateBase():
    def __str__(self) -> str: ...
    def copy(self) -> QuantumGateBase: 
        """
        Create copied instance
        """
    def get_control_index_list(self) -> typing.List[int]: 
        """
        Get control qubit index list
        """
    def get_control_index_value_list(self) -> typing.List[typing.Tuple[int, int]]: 
        """
        Get control qubit pair index value list
        """
    def get_control_value_list(self) -> typing.List[int]: 
        """
        Get control qubit value list
        """
    def get_inverse(self) -> QuantumGateBase: 
        """
        get inverse gate
        """
    def get_matrix(self) -> numpy.ndarray[numpy.complex128, _Shape[m, n]]: 
        """
        Get gate matrix
        """
    def get_name(self) -> str: 
        """
        Get gate name
        """
    def get_target_index_list(self) -> typing.List[int]: 
        """
        Get target qubit index list
        """
    def is_Clifford(self) -> bool: 
        """
        Check this gate is element of Clifford group
        """
    def is_Gaussian(self) -> bool: 
        """
        Check this gate is element of Gaussian group
        """
    def is_Pauli(self) -> bool: 
        """
        Check this gate is element of Pauli group
        """
    def is_commute(self, gate: QuantumGateBase) -> bool: 
        """
        Check this gate commutes with a given gate
        """
    def is_diagonal(self) -> bool: 
        """
        Check the gate matrix is diagonal
        """
    def is_parametric(self) -> bool: 
        """
        Check this gate is parametric gate
        """
    def to_string(self) -> str: 
        """
        to string
        """
    def update_quantum_state(self, state: QuantumStateBase) -> None: 
        """
        Update quantum state
        """
    pass
class ClsNoisyEvolution_fast(QuantumGateBase):
    pass
class ClsOneControlOneTargetGate(QuantumGateBase):
    pass
class ClsOneQubitGate(QuantumGateBase):
    pass
class ClsOneQubitRotationGate(QuantumGateBase):
    pass
class ClsPauliGate(QuantumGateBase):
    pass
class ClsPauliRotationGate(QuantumGateBase):
    pass
class ClsReversibleBooleanGate(QuantumGateBase):
    pass
class ClsStateReflectionGate(QuantumGateBase):
    pass
class ClsTwoQubitGate(QuantumGateBase):
    pass
class QuantumStateBase():
    pass
class GeneralQuantumOperator():
    def __IADD__(self, arg0: PauliOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def __IMUL__(self, arg0: PauliOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def __IMUL__(self, arg0: complex) -> GeneralQuantumOperator: ...
    def __ISUB__(self, arg0: PauliOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def __add__(self, arg0: GeneralQuantumOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def __add__(self, arg0: PauliOperator) -> GeneralQuantumOperator: ...
    def __iadd__(self, arg0: GeneralQuantumOperator) -> GeneralQuantumOperator: ...
    def __imul__(self, arg0: GeneralQuantumOperator) -> GeneralQuantumOperator: ...
    def __init__(self, qubit_count: int) -> None: 
        """
        Constructor
        """
    def __isub__(self, arg0: GeneralQuantumOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def __mul__(self, arg0: GeneralQuantumOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def __mul__(self, arg0: PauliOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def __mul__(self, arg0: complex) -> GeneralQuantumOperator: ...
    def __str__(self) -> str: 
        """
        to string
        """
    @typing.overload
    def __sub__(self, arg0: GeneralQuantumOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def __sub__(self, arg0: PauliOperator) -> GeneralQuantumOperator: ...
    @typing.overload
    def add_operator(self, coef: complex, pauli_string: str) -> None: 
        """
        Add Pauli operator
        """
    @typing.overload
    def add_operator(self, pauli_operator: PauliOperator) -> None: ...
    def add_operator_copy(self, pauli_operator: PauliOperator) -> None: 
        """
        Add Pauli operator
        """
    def add_operator_move(self, pauli_operator: PauliOperator) -> None: 
        """
        Add Pauli operator
        """
    @typing.overload
    def apply_to_state(self, state_to_be_multiplied: QuantumStateBase, dst_state: QuantumStateBase) -> None: 
        """
        Apply observable to `state_to_be_multiplied`. The result is stored into `dst_state`.
        """
    @typing.overload
    def apply_to_state(self, work_state: QuantumStateBase, state_to_be_multiplied: QuantumStateBase, dst_state: QuantumStateBase) -> None: ...
    def copy(self) -> GeneralQuantumOperator: 
        """
        Create copied instance of General Quantum operator class
        """
    def get_expectation_value(self, state: QuantumStateBase) -> complex: 
        """
        Get expectation value
        """
    def get_expectation_value_single_thread(self, state: QuantumStateBase) -> complex: 
        """
        Get expectation value
        """
    def get_qubit_count(self) -> int: 
        """
        Get qubit count
        """
    def get_state_dim(self) -> int: 
        """
        Get state dimension
        """
    def get_term(self, index: int) -> PauliOperator: 
        """
        Get Pauli term
        """
    def get_term_count(self) -> int: 
        """
        Get count of Pauli terms
        """
    def get_transition_amplitude(self, state_bra: QuantumStateBase, state_ket: QuantumStateBase) -> complex: 
        """
        Get transition amplitude
        """
    def is_hermitian(self) -> bool: 
        """
        Get is Herimitian
        """
    pass
class GradCalculator():
    def __init__(self) -> None: ...
    @typing.overload
    def calculate_grad(self, parametric_circuit: ParametricQuantumCircuit, observable: Observable) -> typing.List[complex]: 
        """
        Calculate Grad
        """
    @typing.overload
    def calculate_grad(self, parametric_circuit: ParametricQuantumCircuit, observable: Observable, angles_of_gates: typing.List[float]) -> typing.List[complex]: ...
    pass
class NoiseSimulator():
    def __init__(self, arg0: QuantumCircuit, arg1: QuantumState) -> None: 
        """
        Constructor
        """
    def execute(self, arg0: int) -> typing.List[int]: 
        """
        Sampling & Return result [array]
        """
    def execute_and_get_result(self, arg0: int) -> SimulationResult: 
        """
        Simulate & Return ressult [array of (state, frequency)]
        """
    pass
class Observable(GeneralQuantumOperator):
    def __init__(self, qubit_count: int) -> None: 
        """
        Constructor
        """
    def __str__(self) -> str: 
        """
        to string
        """
    @typing.overload
    def add_operator(self, coef: complex, string: str) -> None: 
        """
        Add Pauli operator
        """
    @typing.overload
    def add_operator(self, pauli_operator: PauliOperator) -> None: ...
    def add_operator_copy(self, pauli_operator: PauliOperator) -> None: 
        """
        Add Pauli operator
        """
    def add_operator_move(self, pauli_operator: PauliOperator) -> None: 
        """
        Add Pauli operator
        """
    @typing.overload
    def add_random_operator(self, operator_count: int) -> None: 
        """
        Add random pauli operator
        """
    @typing.overload
    def add_random_operator(self, operator_count: int, seed: int) -> None: ...
    def apply_to_state(self, work_state: QuantumStateBase, state_to_be_multiplied: QuantumStateBase, dst_state: QuantumStateBase) -> None: 
        """
        Apply observable to `state_to_be_multiplied`. The result is stored into `dst_state`.
        """
    def get_expectation_value(self, state: QuantumStateBase) -> float: 
        """
        Get expectation value
        """
    def get_expectation_value_single_thread(self, state: QuantumStateBase) -> float: 
        """
        Get expectation value
        """
    def get_qubit_count(self) -> int: 
        """
        Get qubit count
        """
    def get_state_dim(self) -> int: 
        """
        Get state dimension
        """
    def get_term(self, index: int) -> PauliOperator: 
        """
        Get Pauli term
        """
    def get_term_count(self) -> int: 
        """
        Get count of Pauli terms
        """
    def get_transition_amplitude(self, state_bra: QuantumStateBase, state_ket: QuantumStateBase) -> complex: 
        """
        Get transition amplitude
        """
    def solve_ground_state_eigenvalue_by_arnoldi_method(self, state: QuantumStateBase, iter_count: int, mu: complex = 0.0) -> complex: 
        """
        Compute ground state eigenvalue by arnoldi method
        """
    def solve_ground_state_eigenvalue_by_lanczos_method(self, state: QuantumStateBase, iter_count: int, mu: complex = 0.0) -> complex: 
        """
        Compute ground state eigenvalue by lanczos method
        """
    def solve_ground_state_eigenvalue_by_power_method(self, state: QuantumStateBase, iter_count: int, mu: complex = 0.0) -> complex: 
        """
        Compute ground state eigenvalue by power method
        """
    pass
class QuantumCircuit():
    def __init__(self, qubit_count: int) -> None: 
        """
        Constructor
        """
    def __str__(self) -> str: 
        """
        to string
        """
    def add_CNOT_gate(self, control: int, target: int) -> None: 
        """
        Add CNOT gate
        """
    def add_CZ_gate(self, control: int, target: int) -> None: 
        """
        Add CNOT gate
        """
    def add_H_gate(self, index: int) -> None: 
        """
        Add Hadamard gate
        """
    def add_P0_gate(self, index: int) -> None: 
        """
        Add projection gate to |0> subspace
        """
    def add_P1_gate(self, index: int) -> None: 
        """
        Add projection gate to |1> subspace
        """
    def add_RX_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-X rotation gate
        """
    def add_RY_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-Y rotation gate
        """
    def add_RZ_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-Z rotation gate
        """
    def add_RotInvX_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-X rotation gate
        """
    def add_RotInvY_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-Y rotation gate
        """
    def add_RotInvZ_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-Z rotation gate
        """
    def add_RotX_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-X rotation gate
        """
    def add_RotY_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-Y rotation gate
        """
    def add_RotZ_gate(self, index: int, angle: float) -> None: 
        """
        Add Pauli-Z rotation gate
        """
    def add_SWAP_gate(self, target1: int, target2: int) -> None: 
        """
        Add SWAP gate
        """
    def add_S_gate(self, index: int) -> None: 
        """
        Add pi/4 phase gate
        """
    def add_Sdag_gate(self, index: int) -> None: 
        """
        Add adjoint of pi/4 phsae gate
        """
    def add_T_gate(self, index: int) -> None: 
        """
        Add pi/8 phase gate
        """
    def add_Tdag_gate(self, index: int) -> None: 
        """
        Add adjoint of pi/8 phase gate
        """
    @staticmethod
    def add_U1_gate(*args, **kwargs) -> typing.Any: 
        """
        Add QASM U1 gate
        """
    @staticmethod
    def add_U2_gate(*args, **kwargs) -> typing.Any: 
        """
        Add QASM U2 gate
        """
    @staticmethod
    def add_U3_gate(*args, **kwargs) -> typing.Any: 
        """
        Add QASM U3 gate
        """
    def add_X_gate(self, index: int) -> None: 
        """
        Add Pauli-X gate
        """
    def add_Y_gate(self, index: int) -> None: 
        """
        Add Pauli-Y gate
        """
    def add_Z_gate(self, index: int) -> None: 
        """
        Add Pauli-Z gate
        """
    @typing.overload
    def add_dense_matrix_gate(self, index: int, matrix: numpy.ndarray[numpy.complex128, _Shape[m, n]]) -> None: 
        """
        Add dense matrix gate
        """
    @typing.overload
    def add_dense_matrix_gate(self, index_list: typing.List[int], matrix: numpy.ndarray[numpy.complex128, _Shape[m, n]]) -> None: ...
    def add_diagonal_observable_rotation_gate(self, observable: Observable, angle: float) -> None: 
        """
        Add diagonal observable rotation gate
        """
    @typing.overload
    def add_gate(self, gate: QuantumGateBase) -> None: 
        """
        Add gate with copy
        """
    @typing.overload
    def add_gate(self, gate: QuantumGateBase, position: int) -> None: ...
    @typing.overload
    def add_multi_Pauli_gate(self, index_list: typing.List[int], pauli_ids: typing.List[int]) -> None: 
        """
        Add multi-qubit Pauli gate
        """
    @typing.overload
    def add_multi_Pauli_gate(self, pauli: PauliOperator) -> None: ...
    @typing.overload
    def add_multi_Pauli_rotation_gate(self, index_list: typing.List[int], pauli_ids: typing.List[int], angle: float) -> None: 
        """
        Add multi-qubit Pauli rotation gate
        """
    @typing.overload
    def add_multi_Pauli_rotation_gate(self, pauli: PauliOperator) -> None: ...
    def add_noise_gate(self, gate: QuantumGateBase, NoiseType: str, NoiseProbability: float) -> None: 
        """
        Add noise gate with copy
        """
    def add_observable_rotation_gate(self, observable: Observable, angle: float, repeat: int) -> None: 
        """
        Add observable rotation gate
        """
    @typing.overload
    def add_random_unitary_gate(self, index_list: typing.List[int]) -> None: 
        """
        Add random unitary gate
        """
    @typing.overload
    def add_random_unitary_gate(self, index_list: typing.List[int], seed: int) -> None: ...
    def add_sqrtX_gate(self, index: int) -> None: 
        """
        Add pi/4 Pauli-X rotation gate
        """
    def add_sqrtXdag_gate(self, index: int) -> None: 
        """
        Add adjoint of pi/4 Pauli-X rotation gate
        """
    def add_sqrtY_gate(self, index: int) -> None: 
        """
        Add pi/4 Pauli-Y rotation gate
        """
    def add_sqrtYdag_gate(self, index: int) -> None: 
        """
        Add adjoint of pi/4 Pauli-Y rotation gate
        """
    def calculate_depth(self) -> int: 
        """
        Calculate depth of circuit
        """
    def copy(self) -> QuantumCircuit: 
        """
        Create copied instance
        """
    def get_gate(self, position: int) -> QuantumGateBase: 
        """
        Get gate instance
        """
    def get_gate_count(self) -> int: 
        """
        Get gate count
        """
    def get_inverse(self) -> QuantumCircuit: 
        """
        get inverse circuit
        """
    def get_qubit_count(self) -> int: 
        """
        Get qubit count
        """
    def merge_circuit(self, circuit: QuantumCircuit) -> None: ...
    def remove_gate(self, position: int) -> None: 
        """
        Remove gate
        """
    def to_string(self) -> str: 
        """
        Get string representation
        """
    @typing.overload
    def update_quantum_state(self, state: QuantumStateBase) -> None: 
        """
        Update quantum state
        """
    @typing.overload
    def update_quantum_state(self, state: QuantumStateBase, start: int, end: int) -> None: ...
    pass
class PauliOperator():
    def __IMUL__(self, arg0: complex) -> PauliOperator: ...
    def __imul__(self, arg0: PauliOperator) -> PauliOperator: ...
    @typing.overload
    def __init__(self, coef: complex) -> None: 
        """
        Constructor
        """
    @typing.overload
    def __init__(self, pauli_string: str, coef: complex) -> None: ...
    @typing.overload
    def __mul__(self, arg0: PauliOperator) -> PauliOperator: ...
    @typing.overload
    def __mul__(self, arg0: complex) -> PauliOperator: ...
    def add_single_Pauli(self, index: int, pauli_type: int) -> None: 
        """
        Add Pauli operator to this term
        """
    def change_coef(self, new_coef: complex) -> None: 
        """
        Change coefficient
        """
    def copy(self) -> PauliOperator: 
        """
        Create copied instance of Pauli operator class
        """
    def get_coef(self) -> complex: 
        """
        Get coefficient of Pauli term
        """
    def get_expectation_value(self, state: QuantumStateBase) -> complex: 
        """
        Get expectation value
        """
    def get_expectation_value_single_thread(self, state: QuantumStateBase) -> complex: 
        """
        Get expectation value
        """
    def get_index_list(self) -> typing.List[int]: 
        """
        Get list of target qubit indices
        """
    def get_pauli_id_list(self) -> typing.List[int]: 
        """
        Get list of Pauli IDs (I,X,Y,Z) = (0,1,2,3)
        """
    def get_pauli_string(self) -> str: 
        """
        Get pauli string
        """
    def get_transition_amplitude(self, state_bra: QuantumStateBase, state_ket: QuantumStateBase) -> complex: 
        """
        Get transition amplitude
        """
    pass
class ParametricQuantumCircuit(QuantumCircuit):
    def __init__(self, qubit_count: int) -> None: 
        """
        Constructor
        """
    def __str__(self) -> str: 
        """
        to string
        """
    @typing.overload
    def add_gate(self, gate: QuantumGateBase) -> None: 
        """
        Add gate
        """
    @typing.overload
    def add_gate(self, gate: QuantumGateBase, position: int) -> None: ...
    def add_parametric_RX_gate(self, index: int, angle: float) -> None: 
        """
        Add parametric Pauli-X rotation gate
        """
    def add_parametric_RY_gate(self, index: int, angle: float) -> None: 
        """
        Add parametric Pauli-Y rotation gate
        """
    def add_parametric_RZ_gate(self, index: int, angle: float) -> None: 
        """
        Add parametric Pauli-Z rotation gate
        """
    @typing.overload
    def add_parametric_gate(self, gate: QuantumGate_SingleParameter) -> None: 
        """
        Add parametric gate
        """
    @typing.overload
    def add_parametric_gate(self, gate: QuantumGate_SingleParameter, position: int) -> None: ...
    def add_parametric_multi_Pauli_rotation_gate(self, index_list: typing.List[int], pauli_ids: typing.List[int], angle: float) -> None: 
        """
        Add parametric multi-qubit Pauli rotation gate
        """
    def backprop(self, obs: GeneralQuantumOperator) -> typing.List[float]: 
        """
        Do backprop
        """
    def backprop_inner_product(self, state: QuantumState) -> typing.List[float]: 
        """
        Do backprop with innder product
        """
    def copy(self) -> ParametricQuantumCircuit: 
        """
        Create copied instance
        """
    def get_parameter(self, index: int) -> float: 
        """
        Get parameter
        """
    def get_parameter_count(self) -> int: 
        """
        Get parameter count
        """
    def get_parametric_gate_position(self, index: int) -> int: 
        """
        Get parametric gate position
        """
    def merge_circuit(self, circuit: ParametricQuantumCircuit) -> None: 
        """
        Merge another ParametricQuantumCircuit
        """
    def remove_gate(self, position: int) -> None: 
        """
        Remove gate
        """
    def set_parameter(self, index: int, parameter: float) -> None: 
        """
        Set parameter
        """
    pass
class QuantumCircuitSimulator():
    def __init__(self, circuit: QuantumCircuit, state: QuantumStateBase) -> None: 
        """
        Constructor
        """
    def copy_state_from_buffer(self) -> None: 
        """
        Copy buffer to state
        """
    def copy_state_to_buffer(self) -> None: 
        """
        Copy state to buffer
        """
    def get_expectation_value(self, observable: Observable) -> complex: 
        """
        Get expectation value
        """
    def get_gate_count(self) -> int: 
        """
        Get gate count
        """
    @typing.overload
    def initialize_random_state(self) -> None: 
        """
        Initialize state with random pure state
        """
    @typing.overload
    def initialize_random_state(self, seed: int) -> None: ...
    def initialize_state(self, arg0: int) -> None: 
        """
        Initialize state
        """
    def simulate(self) -> None: 
        """
        Simulate circuit
        """
    def simulate_range(self, start: int, end: int) -> None: 
        """
        Simulate circuit
        """
    def swap_state_and_buffer(self) -> None: 
        """
        Swap state and buffer
        """
    pass
class ClsNoisyEvolution(QuantumGateBase):
    pass
class QuantumGateDiagonalMatrix(QuantumGateBase):
    pass
class QuantumGateMatrix(QuantumGateBase):
    def add_control_qubit(self, index: int, control_value: int) -> None: 
        """
        Add control qubit
        """
    def multiply_scalar(self, value: complex) -> None: 
        """
        Multiply scalar value to gate matrix
        """
    pass
class QuantumGateSparseMatrix(QuantumGateBase):
    pass
class QuantumGate_Adaptive(QuantumGateBase):
    pass
class QuantumGate_CP(QuantumGateBase):
    pass
class QuantumGate_CPTP(QuantumGateBase):
    """
    QuantumGate_Instrument
    """
    pass
class QuantumGate_Probabilistic(QuantumGateBase):
    """
    QuantumGate_ProbabilisticInstrument
    """
    def get_cumulative_distribution(self) -> typing.List[float]: 
        """
        get_cumulative_distribution
        """
    def get_distribution(self) -> typing.List[float]: 
        """
        get_distribution
        """
    def get_gate_list(self) -> typing.List[QuantumGateBase]: 
        """
        get_gate_list
        """
    def optimize_ProbablisticGate(self) -> None: 
        """
        optimize_ProbablisticGate
        """
    pass
class QuantumGate_SingleParameter(QuantumGateBase):
    def copy(self) -> QuantumGate_SingleParameter: 
        """
        Create copied instance
        """
    def get_parameter_value(self) -> float: 
        """
        Get parameter value
        """
    def set_parameter_value(self, value: float) -> None: 
        """
        Set parameter value
        """
    pass
class QuantumState(QuantumStateBase):
    def __init__(self, qubit_count: int) -> None: 
        """
        Constructor
        """
    def __str__(self) -> str: 
        """
        to string
        """
    def add_state(self, state: QuantumStateBase) -> None: 
        """
        Add state vector to this state
        """
    def allocate_buffer(self) -> QuantumState: 
        """
        Allocate buffer with the same size
        """
    def copy(self) -> QuantumState: 
        """
        Create copied instance
        """
    def get_amplitude(self, comp_basis: int) -> complex: 
        """
        Get Amplitude of a specified computational basis
        """
    def get_classical_value(self, index: int) -> int: 
        """
        Get classical value
        """
    def get_device_name(self) -> str: 
        """
        Get allocated device name
        """
    def get_entropy(self) -> float: 
        """
        Get entropy
        """
    def get_marginal_probability(self, measured_values: typing.List[int]) -> float: 
        """
        Get merginal probability for measured values
        """
    def get_qubit_count(self) -> int: 
        """
        Get qubit count
        """
    def get_squared_norm(self) -> float: 
        """
        Get squared norm
        """
    def get_vector(self) -> numpy.ndarray[numpy.complex128, _Shape[m, 1]]: 
        """
        Get state vector
        """
    def get_zero_probability(self, index: int) -> float: 
        """
        Get probability with which we obtain 0 when we measure a qubit
        """
    @typing.overload
    def load(self, state: QuantumStateBase) -> None: 
        """
        Load quantum state vector
        """
    @typing.overload
    def load(self, state: typing.List[complex]) -> None: ...
    def multiply_coef(self, coef: complex) -> None: 
        """
        Multiply coefficient to this state
        """
    def multiply_elementwise_function(self, func: typing.Callable[[int], complex]) -> None: 
        """
        Multiply elementwise function
        """
    def normalize(self, squared_norm: float) -> None: 
        """
        Normalize quantum state
        """
    @typing.overload
    def sampling(self, sampling_count: int) -> typing.List[int]: 
        """
        Sampling measurement results
        """
    @typing.overload
    def sampling(self, sampling_count: int, random_seed: int) -> typing.List[int]: ...
    @typing.overload
    def set_Haar_random_state(self) -> None: 
        """
        Set Haar random state
        """
    @typing.overload
    def set_Haar_random_state(self, seed: int) -> None: ...
    def set_classical_value(self, index: int, value: int) -> None: 
        """
        Set classical value
        """
    def set_computational_basis(self, comp_basis: int) -> None: 
        """
        Set state to computational basis
        """
    def set_zero_state(self) -> None: 
        """
        Set state to |0>
        """
    def to_string(self) -> str: 
        """
        to string
        """
    pass
class DensityMatrix(QuantumStateBase):
    def __init__(self, qubit_count: int) -> None: 
        """
        Constructor
        """
    def __str__(self) -> str: 
        """
        to string
        """
    def add_state(self, state: QuantumStateBase) -> None: 
        """
        Add state vector to this state
        """
    def allocate_buffer(self) -> DensityMatrix: 
        """
        Allocate buffer with the same size
        """
    def copy(self) -> DensityMatrix: 
        """
        Create copied insntace
        """
    def get_classical_value(self, index: int) -> int: 
        """
        Get classical value
        """
    def get_device_name(self) -> str: 
        """
        Get allocated device name
        """
    def get_entropy(self) -> float: 
        """
        Get entropy
        """
    def get_marginal_probability(self, measured_values: typing.List[int]) -> float: 
        """
        Get merginal probability for measured values
        """
    def get_matrix(self) -> numpy.ndarray[numpy.complex128, _Shape[m, n]]: 
        """
        Get density matrix
        """
    def get_qubit_count(self) -> int: 
        """
        Get qubit count
        """
    def get_squared_norm(self) -> float: 
        """
        Get squared norm
        """
    def get_zero_probability(self, index: int) -> float: 
        """
        Get probability with which we obtain 0 when we measure a qubit
        """
    @typing.overload
    def load(self, state: QuantumStateBase) -> None: 
        """
        Load quantum state vector or density matrix
        """
    @typing.overload
    def load(self, state: numpy.ndarray[numpy.complex128, _Shape[m, n]]) -> None: ...
    @typing.overload
    def load(self, state: typing.List[complex]) -> None: ...
    def multiply_coef(self, coef: complex) -> None: 
        """
        Multiply coefficient to this state
        """
    def normalize(self, squared_norm: float) -> None: 
        """
        Normalize quantum state
        """
    @typing.overload
    def sampling(self, sampling_count: int) -> typing.List[int]: 
        """
        Sampling measurement results
        """
    @typing.overload
    def sampling(self, sampling_count: int, random_seed: int) -> typing.List[int]: ...
    @typing.overload
    def set_Haar_random_state(self) -> None: 
        """
        Set Haar random state
        """
    @typing.overload
    def set_Haar_random_state(self, seed: int) -> None: ...
    def set_classical_value(self, index: int, value: int) -> None: 
        """
        Set classical value
        """
    def set_computational_basis(self, comp_basis: int) -> None: 
        """
        Set state to computational basis
        """
    def set_zero_state(self) -> None: 
        """
        Set state to |0>
        """
    def to_string(self) -> str: 
        """
        to string
        """
    pass
class SimulationResult():
    def get_count(self) -> int: 
        """
        get state count
        """
    def get_frequency(self, arg0: int) -> int: 
        """
        get state frequency
        """
    def get_state(self, arg0: int) -> QuantumState: 
        """
        get state
        """
    pass
def StateVector(arg0: int) -> QuantumState:
    """
    StateVector
    """
def to_general_quantum_operator(gate: QuantumGateBase, qubits: int, tol: float) -> GeneralQuantumOperator:
    pass



from typing import Tuple

import qulacs_core

def create_quantum_operator_from_openfermion_file(arg0: str) -> qulacs_core.GeneralQuantumOperator: ...
def create_quantum_operator_from_openfermion_text(arg0: str) -> qulacs_core.GeneralQuantumOperator: ...
def create_split_quantum_operator(arg0: str) -> Tuple[qulacs_core.GeneralQuantumOperator,qulacs_core.GeneralQuantumOperator]: ...



from __future__ import annotations
import qulacs_core.circuit
import typing
import qulacs_core

__all__ = [
    "QuantumCircuitOptimizer"
]


class QuantumCircuitOptimizer():
    def __init__(self) -> None: 
        """
        Constructor
        """
    def merge_all(self, circuit: qulacs_core.QuantumCircuit) -> qulacs_core.QuantumGateMatrix: ...
    def optimize(self, circuit: qulacs_core.QuantumCircuit, block_size: int) -> None: 
        """
        Optimize quantum circuit
        """
    def optimize_light(self, circuit: qulacs_core.QuantumCircuit) -> None: 
        """
        Optimize quantum circuit with light method
        """
    pass



from typing import Tuple

import qulacs_core

def create_observable_from_openfermion_file(file_path: str) -> qulacs_core.Observable: ...
def create_observable_from_openfermion_text(text: str) -> qulacs_core.Observable: ...
def create_split_observable(arg0: str) -> Tuple[qulacs_core.Observable,qulacs_core.Observable]: ...
